/* Copyright 2000 Servaas Vandenberghe, Leuven, Belgium, All rights reserved.
 This software is licensed under the GNU GENERAL PUBLIC LICENSE, version 2, or
 a later version at your option.

 Redistribution and use of this source code, with or without modification, is
 permitted provided that the following condition is met:

 1. Redistributions of this source code must retain the above copyright
    notice, this condition, and the following disclaimer.

  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
  EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
  ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>   /* getopt */
#include <stdio.h>
#include <string.h>
#include <errno.h> 

#ifndef VERBOSE
#define VERBOSE 0
#endif
#define PROGLEN 1025
#ifndef VERSION
#define VERSION "0" 
#endif
static char progname[PROGLEN]="toascii";
static char version[]= VERSION ", " __DATE__ ", " __TIME__ ;
static char cname[]="toascii";

static void help(void) {
  printf("usage: %s [options] [from-file [to-file]] \n", progname);
  printf("%s substitutes control characters by spaces and replaces ",cname);
  printf("cr,lf[,ff] by lf. \n");
  printf("-B   remove 0x80 from the input stream \n"
         "-b   remove control characters from the stream, except for 0x80 \n"
	 "-i   pass iso-8859 range [0xA0..0xFF] \n"
	 "-F   flush the output stream per processed character \n"
	 "-p   preserve control characters; do not substitute by spaces \n"
	 "-R   reverse the cr,lf processing; replace lf by cr,lf \n"
	 "-W   word to iso-8859 map \n"
	 "-w   word perfect 5 to iso-8859 map \n"
	 );
}

/*
 *			M A I N
 *
 *	This routine opens the file and reads date and attributes.
 * wp:   in   out ref      char
 *       08   E0  agrave   
 *       09   E2  acirc    
 *       0C   E7  ccedille 
 *       0F   E9  eacute   
 *       11   E8  egrave   
 *       12   EA  ecirc    
 *       1F   F9  ugrave   
 *       80   20  space
 *       84   2D  dash
 *       CC   0A  nl
 *
 * wp double octet characters (between xF0 . . xF0):
 *       in      out  ref           char
 *       1C 04   27   single quote  '
 *       33 01   EE   icirc         
 *       3D 01   F4   ocirc         
 *       45 01   FB   ucirc         
 */
#define DEBUG 0
#define BUFSIZE 128
#define WP_MAP 13
int main(int argc, char *argv[]) {
  char where[]="toascii main";
  FILE *fpout=stdout, *fpin=stdin;
  char *fromfile=NULL, *tofile=NULL;
  unsigned ins, inread, outwrite;
  unsigned char inbuf[BUFSIZE], *inpnt, inprev;
  int err=0;
  char newlines[]="\x0d\x0a\x0c";

  /* word perfect 5 map */
  unsigned char wpin[] ="\x08\x09\x0C\x0F\x11\x12\x1F\x80\x84\xCC";
  unsigned char wpout[]="\xE0\xE2\xE7\xE9\xE8\xEA\xF9\x20\x2D\x0A";
  unsigned char *wp_in[WP_MAP+1]={
    "\xF0\x1C\x04\xF0", 
    "\xF0\x33\x01\xF0", /*  */
    "\xF0\x3D\x01\xF0", /*  */
    "\xF0\x45\x01\xF0", /*  */
    "\xF0\x20\x04\xF0", /* bold on */
    "\xF0\x1F\x04\xF0", /* bold off */
    "\xD0\x01",         /* auto NL */
    "\xD0\x04",         /* hard NL */
    "\xD0\x06",         /* page end FF CR */
    "\xD0\x14",         /* line break after hyphen (dash) */
    "\xD4\x5F",         /* interword info, unbreakable space ? */
    "\xDD\x0B",         /* start of document */
    "\xE0\x11",         /* tabbing indent information (before sentence) */
    NULL };
  unsigned wp_inlen[WP_MAP] ={ 4, 4,4,4, 4,4,  2,  2,  2,  2,  2,  2,  2};
  unsigned char *wp_off[WP_MAP]={ 
    NULL, NULL,NULL,NULL, NULL,NULL, 
    "\x00\xD0", "\x00\xD0", "\x00\xD0", "\x00\xD0", 
    "\x00\xD4", "\x00\xDD", "\x00\xE0" };
  unsigned wp_offpos[WP_MAP]={ 0, 0,0,0, 0,0, 19, 19, 19, 19, 14,  9, 10};
  unsigned wp_offlen[WP_MAP]={ 0, 0,0,0, 0,0,  2,  2,  2,  2,  2,  2,  2};

  unsigned char *wp_out[WP_MAP]={ 
    "'", "\xEE", "\xF4", "\xFB", "*-", "-*",
    "\n", "\n\n", "\n", "\n", 
    "", "\n", "  " };
  unsigned wp_outlen[WP_MAP]={ 1, 1,1,1, 2,2,  1,  2,  1,  1,  0,  1,  2};

  /* MS word map */
  char wordin[] ="\x91\x92";
  char wordout[]="\x27\x27";

  unsigned char *xlatin=NULL, *xlatout, 
    **xlat_in=NULL, **xlat_out, **xlat_off=NULL;
  unsigned *xlat_inlen, *xlat_outlen, *xlat_offpos, *xlat_offlen;

  /* options */
  int opt, files;
  extern char *optarg;   /* pointer to the argument of the option */
  extern int optind;     /*, opterr, optopt; */
  int opt_noblank=0;     /* do not substitute a ctrl char by a space */
  int opt_noxblank=0;    /* do not substitute 0x80 by a space */
  int opt_pres=0;        /*   0x80 is a breakable space in word perfect */
  int opt_todos=0;       /* map lf to cr,lf */
  int opt_flush=0;
  int opt_word=0;        /* word to iso-8859 char mapping */
  int opt_wp=0;          /* wp5 to iso-8859 char mapping */
  int opt_iso=0;         /* pass iso-8859 range 0xA0..0xFF */

  optarg=NULL;  /* init getopt */
  optind=0;

  if(argc) strncpy(progname,argv[0],PROGLEN-1);
  progname[PROGLEN-1]=0;

  while ( (opt=getopt(argc, argv, "BbiFhpRwWV")) != EOF ) {
    int this_optind=optind-1;
    switch(opt) {
    case 'B':
      opt_noxblank^=1;
      break;
    case 'b':
      opt_noblank^=1;
      break;
    case 'i':
      opt_iso^=1;
      break;
    case 'F':
      opt_flush^=1;
      break;
    case 'p':
      opt_pres^=1;
      break;
    case 'h':
      help(); goto end_error;
    case 'R':
      opt_todos^=1;
      break;
    case 'W':
      opt_word^=1;
      opt_iso^=1;
      break;
    case 'w':
      opt_wp^=1;
      break;
    case 'V':
      printf("%s %s\n", cname,version); goto end_error;
    case ':':
      fprintf(stderr,"%s: Expected parameter at option %i\n",where,
	      this_optind);
      err=10; goto end_error;	
    case '?':
      fprintf(stderr,"%s: Getopt returned error at option %i\n",
	      where,this_optind);
      err=11; goto end_error;
      break;
    default:
      fprintf(stderr,"%s: ?? getopt returned character code 0x%x\n",where,opt);
      err=12; goto end_error;
    }
  }

  /*** check if file name present ***/
  files=argc-optind;
  if (files>=1) {
    fromfile=argv[optind];
    if(files>=2) tofile=argv[optind+1];
    if(files>2) {
      fprintf(stderr,"%s: Extra garbage at command line\n",where);
      err=1; goto end_error;
    }
  }

  if(fromfile && strlen(fromfile)) {  /* open the fromfile */
    if(!(fpin=fopen(fromfile, "r"))) {
      fprintf(stderr,"%s: fopen %s: %i %s\n",where,fromfile,errno,
	      strerror(errno));
      err=2; goto end_error;
    } 
  } /* if(fromfile) */

  if(tofile && strlen(tofile)) {  /* open the tofile */
    if(fromfile && !strcmp(fromfile,tofile)) {
      fprintf(stderr,"%s: fopen %s: equal files\n",where,tofile);
      err=3; goto end_error;
    }
    if(!(fpout=fopen(tofile, "w"))) {
      fprintf(stderr,"%s: fopen %s: %i %s\n",where,tofile,errno,
	      strerror(errno));
      err=4; goto end_error;
    } 
  } /* if(tofile) */

  if(opt_wp) {
    xlatin=wpin;
    xlatout=wpout;
    xlat_in=wp_in;
    xlat_inlen=wp_inlen;
    xlat_off=wp_off;  
    xlat_offpos=wp_offpos;
    xlat_offlen=wp_offlen;
    xlat_out=wp_out;
    xlat_outlen=wp_outlen;
  }
  if(opt_word) {
    xlatin=wordin;
    xlatout=wordout;
  }

  /*** process ***/
  inprev=0;
  ins=inread=fread(inbuf, sizeof(inbuf[0]), BUFSIZE, fpin);
  inpnt=inbuf;
  for(; ins; ) { /* inpnt[0..ins-1] needs processing */
    unsigned char outc[4]={*inpnt, ' ', ' ', ' '};
    unsigned outs=1, inrem=0;

    /* printf("*%02X*", bytein); */
    if(*inpnt==0) {
      inrem=1;
      if(!opt_pres) outc[0]=' ';
      if(opt_noblank) outs=0;
    } 

    if(!inrem && xlatin) { /* xlat single char */
      unsigned char *tmp;
      if((tmp=strchr(xlatin, *inpnt))) { 
	inrem=1;
	outc[0]=xlatout[tmp-xlatin];
      }
    }

    if(!inrem && xlat_in) { /* xlat string */
      unsigned u;
      for(u=0; !inrem && xlat_in[u]; u++) 
	if(ins>=xlat_inlen[u] && !strncmp(inpnt, xlat_in[u], xlat_inlen[u])) {
	  if(xlat_off && xlat_off[u]) {
	    unsigned opos=xlat_offpos[u], olen=xlat_offlen[u],
	      inremtot=opos+olen;
#if DEBUG
	    printf("*  opos=%u olen=%u \n",opos,olen);
#endif
	    if(ins>=inremtot && !strncmp(&inpnt[opos], xlat_off[u], olen))
	      inrem=inremtot;

	  } else {
	    inrem=xlat_inlen[u];
	  }
	}

      if(inrem) { /* xlat success on u-1 */
	u--;
	strncpy(outc, xlat_out[u], xlat_outlen[u]);
	outs=xlat_outlen[u];
#if DEBUG
	printf("** outs=%u u=%u \n",outs,u);
#endif
      }
    } /* xlat string */

    if(!inrem && strchr(newlines, *inpnt)) { /* got cr,lf,ff */
      inrem=1;
      outs=0;
      if(inprev==0 || !strchr(newlines, inprev) || inprev==*inpnt) {
	/* prev char was no cr,lf,ff  or  equal */
	/* removes  cr lf  or  lf cr  sequence */
	if(opt_todos) outc[outs++]='\x0d';
	outc[outs++]='\x0a';
      }
    }

    if(!inrem) {
      if(*inpnt==128) {
	if(!opt_pres) outc[0]=' ';
	if(opt_noxblank) outs=0;
      } else if(*inpnt<32 || (*inpnt>=127 && (!opt_iso || *inpnt<160))) {
	if(!opt_pres) outc[0]=' ';
	if(opt_noblank) outs=0;
      }
    }

    if(!inrem) inrem=1;

    if(outs) {
      if((outwrite=fwrite(outc, sizeof(outc[0]), outs, fpout))<outs) break;
      if(opt_flush) 
	if(fflush(fpout)) break;
    }

    inprev=*inpnt;
    ins  -=inrem;
    inpnt =&inpnt[inrem];

    /* fill inbuf[] buffer */
    if(inread && ins<=BUFSIZE/2) {
      unsigned infree=BUFSIZE-ins;

      memmove(inbuf, inpnt, sizeof(inbuf[0])*ins);
      inpnt=inbuf;

      inread=fread(&inbuf[ins], sizeof(inbuf[0]), infree, fpin);
      ins+=inread;
    }
  } /* for */

  if(ferror(fpin)) {
    fprintf(stderr,"%s: fread: %i %s\n",where,errno,strerror(errno));
    err=10;
  }
  if(ferror(fpout)) { /* error on write */
    fprintf(stderr,"%s: fwrite: %i %s\n",where,errno,strerror(errno));
    err=11;
  }

end_error:
  if(fpout && fpout!=stdout) 
    if(fclose(fpout)) {
      fprintf(stderr,"%s: fclose: %i %s\n",where,errno,strerror(errno));
      if(!err) err=98;
    }

  if(fpin && fpin!=stdin) 
    if(fclose(fpin)) {
      fprintf(stderr,"%s: fclose: %i %s\n",where,errno,strerror(errno));
      if(!err) err=99;
    }

  return(err);
}
#undef WP_MAP
#undef BUFSIZE
#undef DEBUG

#undef VERBOSE
#undef PROGLEN
#undef VERSION
